### Importieren der erforderlichen Python-Pakete
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import sympy as smp
from sympy.matrices import Matrix, zeros # auch ones, eye, randMatrix
import copy
import csv
import math
from scipy.interpolate import interp1d
import time

### Simulationseinstellungen
Ausgaben_schreiben = False                                           # Sollen die Variablen in Jupyter Notebook ausgegeben werden?
Physikalische_Nichtlinearitaet_beruecksichtigen = True               # Nichtlineares Werkstoffverhalten berücksichtigen?
Steifigkeitsmatrix_neu_berechnen = True                              # Soll die Steifigkeitsmatrix immer wieder neu berechnet werden?
Abbruchkriterium_Kraftresiduum = 1.0e-10                             # Abbruchkriterium Kraftresiduum
Maximale_Anzahl_Simulationen = 300                                   # Oder Abbruch nach Erreichen der maximalen Anzahl an Simulationen

### Soll noch in Einlesedatei
A_in_sqmm = 100

#Achtung: Projektname in Package auskommentieren, wenn Inputfile_aus_App
Inputfile_aus_App = False
Projektnummer = 631406

if (Inputfile_aus_App == True): 
    Projektname = str(Projektnummer)
    from lsdt_strucsim_convert import *
    convert_traglast_app(Projektnummer)
else:       
    ### Projektname - Achtung: muss mit Dateinamen übereinstimmen                 
    Projektname = 'lecture'                                              
    #Projektname = 'bridge'                                              
    #Projektname = 'chapter2'                                              
    #Projektname = 'chapter4'                                              
    #Projektname = 'chapter5'                                              
    #Projektname = 'chapter6'                                              
    #Projektname = 'chapter7'                                              
    #Projektname = 'chapter8'                                              
    #Projektname = 'chapter9'                                              
                               
Datei_Knoten = '../input_csv/'+Projektname+'_Knoten.csv'                    # Datei mit Knotenkoordinaten
Datei_Verbindungen = '../input_csv/'+Projektname+'_Verbindungen.csv'        # Datei mit Knotenverbindungen
Datei_Kraefte = '../input_csv/'+Projektname+'_Kraefte.csv'                  # Datei mit Info über externe Kräfte
Datei_Lager = '../input_csv/'+Projektname+'_Lager.csv'                      # Datei mit Info über Lager
Datei_Material = '../input_csv/'+Projektname+'_Material.csv'                # Datei mit Info über Material
Datei_Spannung_Dehnung = '../input_csv/'+Projektname+'_sigma_epsilon.csv'   # Spannungs-Dehnungskurven aller Materialien in einer Datei

polynomgrad = 1                                                             # Polynomgrad der Ansatzfunktion

sigma_max_in_MPa = 200                                                      # maximal zulässige Spannung des Werkstoffes
Sicherheitsfaktor_Auslegung = 1                                             # Sicherheitsfaktor zur maximalen Spannung
Darstellungsfaktor_Verformung = 10                                          # Faktor um den die Verformung in der Darstellung vergrößert wird

### Einstellungen für Erstellung der Diagramme

plt.rcParams['figure.dpi']  = 300
plt.rcParams['savefig.dpi'] = 300
configuration_transparent = False

#Definition Datenlogger
class class_datenlogger:
    def __init__(self):        
        self.Iterationsnummer = []
        self.Kraftresiduum = []
    
#Definion Materialklasse
class class_material:
    def __init__(self, Material_id):
        self.Material_id = Material_id      # Material_id
        self.Werte_vorhanden = False        # Sind Werte eingetragen oder ist dieses Material leer?
        self.epsilon = []                   # Array Verzerrungen
        self.sigma = []                     # Array Spannungen        
        #self.f_epsilon_sigma               # Interpolationsfunktion Dehnung -> Spannung      

#Definion Materialbibliotheksklasse
class class_bibliothek_materialien:
    def __init__(self):
        self.Material = []
        self.Material.append(class_material(0))        
        self.Material.append(class_material(1))        
        self.Material.append(class_material(2))        
        self.Material.append(class_material(3))        
        self.Material.append(class_material(4))        
        self.Material.append(class_material(5))        

### Definition Knotenklasse
class class_knoten:
    def __init__(self):         
        self.Anzahl_Knoten = 0                        # Gesamtanzahl der Knoten        
        self.id_Knoten = []                           # Knotennummer
        self.pos_x_in_mm = []                         # Knotenkoordinate x
        self.pos_y_in_mm = []                         # Knotenkoordinate y
        self.disp_x_in_mm = []                        # Knotenverschiebung in x-Richtung
        self.disp_y_in_mm = []                        # Knotenverschiebung in y-Richtung
        self.delta_disp_x_in_mm = []                  # Knotenverschiebung aufgrund des Kraftungleichgewichtes in x-Richtung
        self.delta_disp_y_in_mm = []                  # Knotenverschiebung aufgrund des Kraftungleichgewichtes in y-Richtung
        self.F_x_stern_in_N = []                      # Äußere Knotenkraft in x-Richtung
        self.F_y_stern_in_N = []                      # Äußere Knotenkraft in y-Richtung        
        self.F_x_reactio_in_N = []                    # Innere Knotenkraft in x-Richtung
        self.F_y_reactio_in_N = []                    # Innere Knotenkraft in y-Richtung  
        self.delta_F_x_in_N = []                      # Kraftungleichgewicht aufgrund des nichtlinearen Verhaltens in x-Richtung        
        self.delta_F_y_in_N = []                      # Kraftungleichgewicht aufgrund des nichtlinearen Verhaltens in y-Richtung        
        self.Lager_x = []                             # Lager in x-Richtung
        self.Lager_y = []                             # Lager in y-Richtung               
        
### Definition Elementklasse
class class_elemente:
    def __init__(self):                 
        self.Anzahl_Elemente = 0             # Gesamtanzahl der Elemente        
        self.id_Element = []                 # Elementnummer
        self.Polynomgrad = polynomgrad       # Polynomgrad des Elementes
        self.id_Knoten_i = []                # Knotennummer des Knotens i
        self.id_Knoten_j = []                # Knotennummer des Knotens j        
        self.C = []                          # Cosinus(Winkel des Elementes)
        self.S = []                          # Sinus(Winkel des Elementes)        
        self.Material_id = []                # Matrerial_id des zugeordneten Materials        
        self.epsilon = []                    # Verzerrung des Elementes, aus den Verschiebungen berechnet
        self.F_res_in_N = []                 # Stabkraft
        self.F_i_x_in_N = []                 # Stabkraft Knoten i in x-Richtung
        self.F_i_y_in_N = []                 # Stabkraft Knoten i in y-Richtung
        self.F_j_x_in_N = []                 # Stabkraft Knoten j in x-Richtung
        self.F_j_y_in_N = []                 # Stabkraft Knoten j in y-Richtung
        self.sigma_linear_in_MPa = []        # Spannung des Elementes berechnet nach dem E-Modul (lineare Theorie)
        self.sigma_nichtlinear_in_MPa = []   # Spannung des Elementes berechnet nach dem eingelesen Werkstoffverhalten (nichtlineare Theorie)        
        self.E_in_MPa = []                   # Elastizitätsmodul des Elementes
        self.A_in_sqmm = []                  # Querschnittsfläche des Elementes
        self.L_in_mm = []                    # Länge des Elementes        
        self.xi = []                         # xi passend zur Formfunktion
        self.N = []                          # Formfunktion
        self.B = []                          # Ableitung der Formfunktion
        self.KE_normiert = []                # normierte Elementsteifigkeitsmatrix (ohne elementspezifische Informationen A,E,L)
        self.KE = []                         # Elementsteifigkeitsmatrix
        self.tabelle_elemente = []           # Tabelle mit Ausgabewerten       

class class_strucsim_funktionen:
    def __init__(self, bibliothek_materialien, liste_knoten, liste_elemente):
        self.bibliothek_materialien = bibliothek_materialien
        self.liste_knoten = liste_knoten
        self.liste_elemente = liste_elemente
        self.Zeit_Start_Simulation_linear = 0
        self.Zeit_Ende_Simulation_linear = 0
        self.Zeit_Dauer_Simulation_linear = 0
        self.Zeit_Start_Simulation_nichtlinear = 0
        self.Zeit_Ende_Simulation_nichtlinear = 0
        self.Zeit_Dauer_Simulation_nichtlinear = 0
        self.K = []
        self.K_Lager = []
        self.K_Lager_inv = []
        self.F = []
        self.delta_F = []
        self.u = []
        self.delta_u = []        
        self.F_reactio = []        
        self.Residuum_F_K = 0                     # Kraftresiduum der äußeren Kräfte

    ###########################################################################
    #### Ab hier Funktionen zum Einlesen von Werten ###########################
    ###########################################################################
    
    def FEMI_K2_Einlesen_Knotenkoordinaten(self):
        with open(Datei_Knoten,'r') as csvdatei:
            csv_reader_object = csv.reader(csvdatei, delimiter=';')

            zeilennummer = 0
            for row in csv_reader_object:
                if (zeilennummer == 0):
                    #Dies ist die erste Zeile in der Datei mit der Spaltenangabe
                    zeilennummer += 1
                else:                    
                    self.liste_knoten.id_Knoten.append(int(row[0]))
                    self.liste_knoten.pos_x_in_mm.append(float(row[1]))
                    self.liste_knoten.pos_y_in_mm.append(float(row[2]))        
                    self.liste_knoten.disp_x_in_mm.append(0.0)
                    self.liste_knoten.disp_y_in_mm.append(0.0)
                    self.liste_knoten.delta_disp_x_in_mm.append(0.0)
                    self.liste_knoten.delta_disp_y_in_mm.append(0.0)
                    self.liste_knoten.F_x_stern_in_N.append(0.0)           
                    self.liste_knoten.F_y_stern_in_N.append(0.0)                               
                    self.liste_knoten.F_x_reactio_in_N.append(0.0)         
                    self.liste_knoten.F_y_reactio_in_N.append(0.0)                             
                    self.liste_knoten.delta_F_x_in_N.append(0.0)         
                    self.liste_knoten.delta_F_y_in_N.append(0.0)                             
                    self.liste_knoten.Lager_x.append(0)                    #wird später noch eingelesen
                    self.liste_knoten.Lager_y.append(0)                    #wird später noch eingelesen
                    zeilennummer += 1
            
            if (Ausgaben_schreiben == True):
                print(Datei_Knoten + ' eingelesen')
                print(f'Anzahl Datensätze importiert: {zeilennummer-1} \n')    
                print("\n")
                
                print("ID Knoten:")
                print(self.liste_knoten.id_Knoten)
                print("\n")
                
                print("Knotenkoordinaten x [mm]:")
                print(self.liste_knoten.pos_x_in_mm)
                print("\n")
                
                print("Knotenkoordinaten y [mm]:")
                print(self.liste_knoten.pos_y_in_mm)
                print("\n")
        
        #Anzahl der Knoten berechnen
        self.liste_knoten.Anzahl_Knoten = len(self.liste_knoten.id_Knoten)
        
        if (Ausgaben_schreiben == True):
            print("Anzahl der Knoten:")
            print(self.liste_knoten.Anzahl_Knoten)
            print("\n")
    
    def FEMI_K2_Einlesen_Knotenverbindungen(self):
        with open(Datei_Verbindungen,'r') as csvdatei:
            csv_reader_object = csv.reader(csvdatei, delimiter=';')

            zeilennummer = 0
            for row in csv_reader_object:
                if (zeilennummer == 0):
                    #Dies ist die erste Zeile in der Datei mit der Spaltenangabe
                    zeilennummer += 1
                else:
                    self.liste_elemente.id_Element.append(int(row[0]))
                    self.liste_elemente.id_Knoten_i.append(int(row[1]))
                    self.liste_elemente.id_Knoten_j.append(int(row[2]))                            
                    self.liste_elemente.C.append(0.0)
                    self.liste_elemente.S.append(0.0)                    
                    self.liste_elemente.epsilon.append(0.0)
                    self.liste_elemente.F_res_in_N.append(0.0)
                    self.liste_elemente.F_i_x_in_N.append(0.0)
                    self.liste_elemente.F_i_y_in_N.append(0.0)
                    self.liste_elemente.F_j_x_in_N.append(0.0)
                    self.liste_elemente.F_j_y_in_N.append(0.0)
                    self.liste_elemente.sigma_linear_in_MPa.append(0.0)
                    self.liste_elemente.sigma_nichtlinear_in_MPa.append(0.0)                    
                    self.liste_elemente.E_in_MPa.append(0.0)
                    self.liste_elemente.A_in_sqmm.append(A_in_sqmm)
                    self.liste_elemente.L_in_mm.append(0.0) #muss später noch berechnet werden                    
                    self.liste_elemente.KE_normiert.append([])
                    self.liste_elemente.KE.append([])
                    self.liste_elemente.tabelle_elemente.append([])
                    zeilennummer += 1
            
            if (Ausgaben_schreiben == True):
                print(Datei_Knoten + ' eingelesen')
                print(f'Anzahl Datensätze importiert: {zeilennummer-1} \n')   
                print("\n")
                
                print("id_Element:")
                print(self.liste_elemente.id_Element)
                print("\n")
                
                print("id_Knoten_i:")
                print(self.liste_elemente.id_Knoten_i)
                print("\n")
                
                print("id_Knoten_j:")
                print(self.liste_elemente.id_Knoten_j)
                print("\n")
                
                print("Material_id:")
                print(self.liste_elemente.Material_id)
                print("\n")
                
        self.liste_elemente.Anzahl_Elemente = len(self.liste_elemente.id_Element)
        
        if (Ausgaben_schreiben == True):
            print("Anzahl_Elemente:")
            print(self.liste_elemente.Anzahl_Elemente)
            print("\n")
    
    def FEMI_K2_Einlesen_Materialzuordnung(self):
        with open(Datei_Material,'r') as csvdatei:
            csv_reader_object = csv.reader(csvdatei, delimiter=';')

            zeilennummer = 0
            for row in csv_reader_object:
                if (zeilennummer == 0):
                    #Dies ist die erste Zeile in der Datei mit der Spaltenangabe
                    zeilennummer += 1
                else:                    
                    self.liste_elemente.Material_id.append(int(row[1]))                                             
                    zeilennummer += 1
            
            if (Ausgaben_schreiben == True):               
                print("Material_id:")
                print(self.liste_elemente.Material_id)
                print("\n")
                
    def FEMI_K2_Einlesen_SpannungsDehnungsKurven(self):
        with open(Datei_Spannung_Dehnung,'r') as csvdatei:
            csv_reader_object = csv.reader(csvdatei, delimiter=';')

            zeilennummer = 0
            for row in csv_reader_object:
                if (zeilennummer == 0):
                    #Dies ist die erste Zeile in der Datei mit der Spaltenangabe
                    zeilennummer += 1
                else:
                    temp_Material_id = int(row[0])                                                         # Material_id steht in der ersten Spalte      
                    self.bibliothek_materialien.Material[temp_Material_id].epsilon.append(float(row[1]))   # Zuordnung des Punktes epsilon in die Bibliothek                   
                    self.bibliothek_materialien.Material[temp_Material_id].sigma.append(float(row[2]))     # Zuordnung des Punktes sigma in die Bibliothek                   
                    self.bibliothek_materialien.Material[temp_Material_id].Werte_vorhanden = True          # Anzeigen, dass Werte für das Material vorhanden sind
                    zeilennummer += 1
            
            if (Ausgaben_schreiben == True):
                print(Datei_Spannung_Dehnung + ' eingelesen')
                print(f'Anzahl Datensätze importiert: {zeilennummer-1} \n')    
                print("\n")
                
                print("Material[1].epsilon:")
                print(self.bibliothek_materialien.Material[1].epsilon)
                print("\n")
                
                print("Material[1].sigma:")
                print(self.bibliothek_materialien.Material[1].sigma)
                print("\n")
                
                print("Material[2].epsilon:")
                print(self.bibliothek_materialien.Material[2].epsilon)
                print("\n")
                
                print("Material[2].sigma:")
                print(self.bibliothek_materialien.Material[2].sigma)
                print("\n")
        
        if (self.bibliothek_materialien.Material[0].Werte_vorhanden == True):
            self.bibliothek_materialien.Material[0].f_epsilon_sigma = interp1d(self.bibliothek_materialien.Material[0].epsilon, self.bibliothek_materialien.Material[0].sigma)                                                                          
            self.FEMI_K2_Plotten_SpannungsDehnungsDiagramm(0)    
            #Erst mal eine Liste anlegen, die abgeleitet werden kann um E-Modul zu bestimmen
            temp_epsilon_min = min(self.bibliothek_materialien.Material[0].epsilon)
            temp_epsilon_max = max(self.bibliothek_materialien.Material[0].epsilon)
            temp_delta_epsilon = (temp_epsilon_max-temp_epsilon_min)/1000.0
            temp_epsilon_list = np.linspace(temp_epsilon_min, temp_epsilon_max, 1000)
            temp_sigma_list = []            
            for index_list in range(len(temp_epsilon_list)):
                temp_sigma_list.append(self.bibliothek_materialien.Material[0].f_epsilon_sigma(temp_epsilon_list[index_list]))            
            temp_E_list = np.gradient(temp_sigma_list,temp_delta_epsilon)            
            self.bibliothek_materialien.Material[0].f_E = interp1d(temp_epsilon_list, temp_E_list)                                                   
                         
            self.FEMI_K2_Plotten_Ableitung_SpannungsDehnungsDiagramm(0)            
        if (self.bibliothek_materialien.Material[1].Werte_vorhanden == True):        
            self.bibliothek_materialien.Material[1].f_epsilon_sigma = interp1d(self.bibliothek_materialien.Material[1].epsilon, self.bibliothek_materialien.Material[1].sigma)            
            self.FEMI_K2_Plotten_SpannungsDehnungsDiagramm(1)
            #Erst mal eine Liste anlegen, die abgeleitet werden kann um E-Modul zu bestimmen
            temp_epsilon_min = min(self.bibliothek_materialien.Material[1].epsilon)
            temp_epsilon_max = max(self.bibliothek_materialien.Material[1].epsilon)
            temp_delta_epsilon = (temp_epsilon_max-temp_epsilon_min)/1000.0
            temp_epsilon_list = np.linspace(temp_epsilon_min, temp_epsilon_max, 1000)
            temp_sigma_list = []            
            for index_list in range(len(temp_epsilon_list)):
                temp_sigma_list.append(self.bibliothek_materialien.Material[1].f_epsilon_sigma(temp_epsilon_list[index_list]))            
            temp_E_list = np.gradient(temp_sigma_list,temp_delta_epsilon)            
            self.bibliothek_materialien.Material[1].f_E = interp1d(temp_epsilon_list, temp_E_list)

            
            self.FEMI_K2_Plotten_Ableitung_SpannungsDehnungsDiagramm(1)
        if (self.bibliothek_materialien.Material[2].Werte_vorhanden == True):        
            self.bibliothek_materialien.Material[2].f_epsilon_sigma = interp1d(self.bibliothek_materialien.Material[2].epsilon, self.bibliothek_materialien.Material[2].sigma)
            self.FEMI_K2_Plotten_SpannungsDehnungsDiagramm(2)
            #Erst mal eine Liste anlegen, die abgeleitet werden kann um E-Modul zu bestimmen
            temp_epsilon_min = min(self.bibliothek_materialien.Material[2].epsilon)
            temp_epsilon_max = max(self.bibliothek_materialien.Material[2].epsilon)
            temp_delta_epsilon = (temp_epsilon_max-temp_epsilon_min)/1000.0
            temp_epsilon_list = np.linspace(temp_epsilon_min, temp_epsilon_max, 1000)
            temp_sigma_list = []            
            for index_list in range(len(temp_epsilon_list)):
                temp_sigma_list.append(self.bibliothek_materialien.Material[2].f_epsilon_sigma(temp_epsilon_list[index_list]))            
            temp_E_list = np.gradient(temp_sigma_list,temp_delta_epsilon)           
            self.bibliothek_materialien.Material[2].f_E = interp1d(temp_epsilon_list, temp_E_list)

                
            self.FEMI_K2_Plotten_Ableitung_SpannungsDehnungsDiagramm(2)
        if (self.bibliothek_materialien.Material[3].Werte_vorhanden == True):
            self.bibliothek_materialien.Material[3].f_epsilon_sigma = interp1d(self.bibliothek_materialien.Material[3].epsilon, self.bibliothek_materialien.Material[3].sigma)
            self.FEMI_K2_Plotten_SpannungsDehnungsDiagramm(3)
            #Erst mal eine Liste anlegen, die abgeleitet werden kann um E-Modul zu bestimmen
            temp_epsilon_min = min(self.bibliothek_materialien.Material[3].epsilon)
            temp_epsilon_max = max(self.bibliothek_materialien.Material[3].epsilon)
            temp_delta_epsilon = (temp_epsilon_max-temp_epsilon_min)/1000.0
            temp_epsilon_list = np.linspace(temp_epsilon_min, temp_epsilon_max, 1000)
            temp_sigma_list = []            
            for index_list in range(len(temp_epsilon_list)):
                temp_sigma_list.append(self.bibliothek_materialien.Material[3].f_epsilon_sigma(temp_epsilon_list[index_list]))            
            temp_E_list = np.gradient(temp_sigma_list,temp_delta_epsilon)            
            self.bibliothek_materialien.Material[3].f_E = interp1d(temp_epsilon_list, temp_E_list)                             

            
            self.FEMI_K2_Plotten_Ableitung_SpannungsDehnungsDiagramm(3)
        if (self.bibliothek_materialien.Material[4].Werte_vorhanden == True):
            self.bibliothek_materialien.Material[4].f_epsilon_sigma = interp1d(self.bibliothek_materialien.Material[4].epsilon, self.bibliothek_materialien.Material[4].sigma)
            self.FEMI_K2_Plotten_SpannungsDehnungsDiagramm(4)
            #Erst mal eine Liste anlegen, die abgeleitet werden kann um E-Modul zu bestimmen
            temp_epsilon_min = min(self.bibliothek_materialien.Material[4].epsilon)
            temp_epsilon_max = max(self.bibliothek_materialien.Material[4].epsilon)
            temp_delta_epsilon = (temp_epsilon_max-temp_epsilon_min)/1000.0
            temp_epsilon_list = np.linspace(temp_epsilon_min, temp_epsilon_max, 1000)
            temp_sigma_list = []            
            for index_list in range(len(temp_epsilon_list)):
                temp_sigma_list.append(self.bibliothek_materialien.Material[4].f_epsilon_sigma(temp_epsilon_list[index_list]))            
            temp_E_list = np.gradient(temp_sigma_list,temp_delta_epsilon)            
            self.bibliothek_materialien.Material[4].f_E = interp1d(temp_epsilon_list, temp_E_list)                                         

            
            self.FEMI_K2_Plotten_Ableitung_SpannungsDehnungsDiagramm(4)
        if (self.bibliothek_materialien.Material[5].Werte_vorhanden == True):
            self.bibliothek_materialien.Material[5].f_epsilon_sigma = interp1d(self.bibliothek_materialien.Material[5].epsilon, self.bibliothek_materialien.Material[5].sigma)
            self.FEMI_K2_Plotten_SpannungsDehnungsDiagramm(5)
            #Erst mal eine Liste anlegen, die abgeleitet werden kann um E-Modul zu bestimmen
            temp_epsilon_min = min(self.bibliothek_materialien.Material[5].epsilon)
            temp_epsilon_max = max(self.bibliothek_materialien.Material[5].epsilon)
            temp_delta_epsilon = (temp_epsilon_max-temp_epsilon_min)/1000.0
            temp_epsilon_list = np.linspace(temp_epsilon_min, temp_epsilon_max, 1000)
            temp_sigma_list = []            
            for index_list in range(len(temp_epsilon_list)):
                temp_sigma_list.append(self.bibliothek_materialien.Material[5].f_epsilon_sigma(temp_epsilon_list[index_list]))            
            temp_E_list = np.gradient(temp_sigma_list,temp_delta_epsilon)            
            self.bibliothek_materialien.Material[5].f_E = interp1d(temp_epsilon_list, temp_E_list)                                         

                
            self.FEMI_K2_Plotten_Ableitung_SpannungsDehnungsDiagramm(5)       
            
        for i in range(len(self.liste_elemente.E_in_MPa)):                 
            self.liste_elemente.E_in_MPa[i] = self.bibliothek_materialien.Material[self.liste_elemente.Material_id[i]].f_E(0.0)
            
    def FEMI_K7_Einlesen_Lagerungen(self):
        with open(Datei_Lager,'r') as csvdatei:
            csv_reader_object = csv.reader(csvdatei, delimiter=';')

            zeilennummer = 0
            for row in csv_reader_object:
                if (zeilennummer == 0):
                    #Dies ist die erste Zeile in der Datei mit der Spaltenangabe
                    zeilennummer += 1
                else:                        
                    self.liste_knoten.Lager_x[zeilennummer-1] = int(row[1])
                    self.liste_knoten.Lager_y[zeilennummer-1] = int(row[2])
                    zeilennummer += 1
            
            if (Ausgaben_schreiben == True):
                print(Datei_Knoten + ' eingelesen')
                print(f'Anzahl Datensätze importiert: {zeilennummer-1} \n') 
                print("\n")
                
        if (Ausgaben_schreiben == True):
            print("Lager_x:")
            print(self.liste_knoten.Lager_x)
            print("\n")
            
            print("Lager_y:")
            print(self.liste_knoten.Lager_y)
            print("\n")
     
    def FEMI_K7_Einlesen_Aeussere_Knotenkraefte(self):
        with open(Datei_Kraefte,'r') as csvdatei:
            csv_reader_object = csv.reader(csvdatei, delimiter=';')

            zeilennummer = 0
            for row in csv_reader_object:
                if (zeilennummer == 0):
                    #Dies ist die erste Zeile in der Datei mit der Spaltenangabe
                    zeilennummer += 1
                else:            
                    self.liste_knoten.F_x_stern_in_N[zeilennummer-1] = float(row[1])
                    self.liste_knoten.F_y_stern_in_N[zeilennummer-1] = float(row[2])
                    zeilennummer += 1
            
            if (Ausgaben_schreiben == True):
                print(Datei_Knoten + ' eingelesen')
                print(f'Anzahl Datensätze importiert: {zeilennummer-1} \n') 
                
        if (Ausgaben_schreiben == True):
            print("F_x_stern_in_N:")
            print(self.liste_knoten.F_x_stern_in_N)
            print("\n")
            
            print("F_y_stern_in_N:")
            print(self.liste_knoten.F_y_stern_in_N)
            print("\n")
            
    ###########################################################################
    #### Ab hier Funktionen zum Darstellen von Ergebnissen ####################
    ###########################################################################
    
    def FEMI_K2_Plotten_SpannungsDehnungsDiagramm(self, index_material):   
        min_epsilon = min(self.bibliothek_materialien.Material[index_material].epsilon)
        max_epsilon = max(self.bibliothek_materialien.Material[index_material].epsilon)
        
        epsilon_plot = np.linspace(min_epsilon,max_epsilon,100)
        sigma_plot = np.linspace(0,0,100)
        
        for index_eintraege in range(len(epsilon_plot)):
            sigma_plot[index_eintraege] = self.bibliothek_materialien.Material[index_material].f_epsilon_sigma(epsilon_plot[index_eintraege])
                                            
        plt.plot(epsilon_plot,sigma_plot)

        plt.title('Werkstoffverhalten')   
        plt.xlabel('Verzerrung []')
        plt.ylabel('Spannung [MPa]')
        plt.savefig("..\output_bilder\epsilon_sigma_"+str(self.bibliothek_materialien.Material[index_material].Material_id)+".png")                                          
        plt.close()
     
    def FEMI_K2_Plotten_Ableitung_SpannungsDehnungsDiagramm(self, index_material):   
        min_epsilon = min(self.bibliothek_materialien.Material[index_material].epsilon)
        max_epsilon = max(self.bibliothek_materialien.Material[index_material].epsilon)
        
        epsilon_plot = np.linspace(min_epsilon,max_epsilon,1000)
        E_plot = np.linspace(0,0,1000)
        
        for index_eintraege in range(len(epsilon_plot)):
            E_plot[index_eintraege] = self.bibliothek_materialien.Material[index_material].f_E(epsilon_plot[index_eintraege])
                                            
        plt.plot(epsilon_plot,E_plot)

        plt.title('Werkstoffverhalten')   
        plt.xlabel('Verzerrung []')
        plt.ylabel('E-Modul [MPa]')
        plt.savefig("..\output_bilder\epsilon_sigma_diff_"+str(self.bibliothek_materialien.Material[index_material].Material_id)+".png")                                          
        plt.close()
        
    def FEMI_K2_Plotten_Modell(self):
        temp_Knoten_x_min_in_mm = min(self.liste_knoten.pos_x_in_mm)
        temp_Knoten_x_max_in_mm = max(self.liste_knoten.pos_x_in_mm)
        temp_Knoten_y_min_in_mm = min(self.liste_knoten.pos_y_in_mm)
        temp_Knoten_y_max_in_mm = max(self.liste_knoten.pos_y_in_mm)

        for index_elemente in range(self.liste_elemente.Anzahl_Elemente):
                
            temp_Knoten_i_id = self.liste_elemente.id_Knoten_i[index_elemente]            
            temp_Knoten_j_id = self.liste_elemente.id_Knoten_j[index_elemente]          
            
            for index_knoten in range(self.liste_knoten.Anzahl_Knoten):
                if (self.liste_knoten.id_Knoten[index_knoten] == temp_Knoten_i_id):
                    temp_Knoten_i_x_in_mm = self.liste_knoten.pos_x_in_mm[index_knoten]
                    temp_Knoten_i_y_in_mm = self.liste_knoten.pos_y_in_mm[index_knoten]
                if (self.liste_knoten.id_Knoten[index_knoten] == temp_Knoten_j_id):
                    temp_Knoten_j_x_in_mm = self.liste_knoten.pos_x_in_mm[index_knoten]
                    temp_Knoten_j_y_in_mm = self.liste_knoten.pos_y_in_mm[index_knoten]                
            
            plt.plot((temp_Knoten_i_x_in_mm, temp_Knoten_j_x_in_mm), (temp_Knoten_i_y_in_mm, temp_Knoten_j_y_in_mm), color='black', linewidth = 1)

        plt.title('Ausgangsmodell',fontsize = 16)   
        plt.xlabel('Koordinate x [mm]')
        plt.ylabel('Koordinate y [mm]')
        plt.tight_layout()        
        plt.savefig('../output_bilder/' + Projektname + '_Modell.png')          
        plt.close()        
    
    def FEMI_K3_K4_Plotten_Formfunktion(self):    
        xi = smp.symbols("xi")
        temp_xi_plot = np.linspace(-1,1,100)
        temp_N_plot = []
        for index_stuetzstellen in range(len(temp_xi_plot)):
            res = {xi:temp_xi_plot[index_stuetzstellen]}
            temp_N_plot.append(self.N.subs(res))
        temp_N_plot_array = np.squeeze(np.asarray(temp_N_plot))        
        temp_knotenbezeichnung = ['i','j','k','l']
        for index_p in range(self.liste_elemente.Polynomgrad+1):
            plt.plot(temp_xi_plot, temp_N_plot_array[:,index_p], label='Knoten:'+temp_knotenbezeichnung[index_p])       
        ax = plt.gca()
        ax.spines['right'].set_color('none')
        ax.spines['top'].set_color('none')
        ax.spines['bottom'].set_position(('data',0))
        ax.xaxis.set_ticks_position('bottom')
        #plt.suptitle('Formfunktion',fontsize = 16)   
        plt.title('Formfunktion Polynomgrad: '+str(self.liste_elemente.Polynomgrad),fontsize = 16)      
        plt.xlabel(r'$\xi$')
        plt.axis([-1, 1, -0.4, 1.1])        
        plt.ylabel(r'$N(\xi)$')
        plt.tight_layout()
        ax.legend()
        plt.savefig('../output_bilder/' + Projektname + '_Formfunktion.png')        
        plt.close()
        
    def FEMI_K8_Plotten_Verformung(self):
        for index_elemente in range(self.liste_elemente.Anzahl_Elemente):
    
            #Hinweis: Nummerierung fängt bei 1 an, Array fängt aber bei 0 an            
            temp_Knoten_i_id = self.liste_elemente.id_Knoten_i[index_elemente]            
            temp_Knoten_j_id = self.liste_elemente.id_Knoten_j[index_elemente]          
            
            for index_knoten in range(self.liste_knoten.Anzahl_Knoten):
                if (self.liste_knoten.id_Knoten[index_knoten] == temp_Knoten_i_id):
                    temp_Knoten_i_pos_x_in_mm = self.liste_knoten.pos_x_in_mm[index_knoten]
                    temp_Knoten_i_pos_y_in_mm = self.liste_knoten.pos_y_in_mm[index_knoten]
                    temp_Knoten_i_disp_x_in_mm = Darstellungsfaktor_Verformung * self.liste_knoten.disp_x_in_mm[index_knoten]
                    temp_Knoten_i_disp_y_in_mm = Darstellungsfaktor_Verformung * self.liste_knoten.disp_y_in_mm[index_knoten]
                if (self.liste_knoten.id_Knoten[index_knoten] == temp_Knoten_j_id):
                    temp_Knoten_j_pos_x_in_mm = self.liste_knoten.pos_x_in_mm[index_knoten]
                    temp_Knoten_j_pos_y_in_mm = self.liste_knoten.pos_y_in_mm[index_knoten]                
                    temp_Knoten_j_disp_x_in_mm = Darstellungsfaktor_Verformung * self.liste_knoten.disp_x_in_mm[index_knoten]
                    temp_Knoten_j_disp_y_in_mm = Darstellungsfaktor_Verformung * self.liste_knoten.disp_y_in_mm[index_knoten]                
            
            temp_max_disp_x = abs(max(self.liste_knoten.disp_x_in_mm))
            temp_min_disp_x = abs(min(self.liste_knoten.disp_x_in_mm))
            
            if (temp_max_disp_x > temp_min_disp_x):
                temp_max_abs_disp_x = temp_max_disp_x
            else:
                temp_max_abs_disp_x = temp_min_disp_x
                
            temp_max_disp_y = abs(max(self.liste_knoten.disp_y_in_mm))
            temp_min_disp_y = abs(min(self.liste_knoten.disp_y_in_mm))
            
            if (temp_max_disp_y > temp_min_disp_y):
                temp_max_abs_disp_y = temp_max_disp_y
            else:
                temp_max_abs_disp_y = temp_min_disp_y
            
            temp_max_abs_disp_res = np.sqrt(pow(temp_max_abs_disp_x,2)+pow(temp_max_abs_disp_y,2))
            #Unverformter Plot
            plt.plot((temp_Knoten_i_pos_x_in_mm, temp_Knoten_j_pos_x_in_mm), (temp_Knoten_i_pos_y_in_mm, temp_Knoten_j_pos_y_in_mm), color='black', linewidth = 1)
                        
            plt.plot((temp_Knoten_i_pos_x_in_mm+temp_Knoten_i_disp_x_in_mm, temp_Knoten_j_pos_x_in_mm+temp_Knoten_j_disp_x_in_mm), (temp_Knoten_i_pos_y_in_mm+temp_Knoten_i_disp_y_in_mm, temp_Knoten_j_pos_y_in_mm+temp_Knoten_j_disp_y_in_mm), color='blue', linewidth = 2, linestyle = '--')

        plt.suptitle('Verformung',fontsize = 16)   
        plt.title('Darstellungsfaktor = '+str(Darstellungsfaktor_Verformung)+ ', max. Verschiebung = '+str(round(temp_max_abs_disp_res,2))+' mm',fontsize = 12 )   
        plt.xlabel('Koordinate x [m]')
        plt.ylabel('Koordinate y [m]')
        plt.tight_layout()
        plt.savefig('../output_bilder/' + Projektname + '_Verformung_linear.png')
        plt.close()
                
    def FEMI_K9_Plotten_Spannungen(self):    
        for index_elemente in range(self.liste_elemente.Anzahl_Elemente):
        
            #Hinweis: Nummerierung fängt bei 1 an, Array fängt aber bei 0 an            
            temp_Knoten_i_id = self.liste_elemente.id_Knoten_i[index_elemente]            
            temp_Knoten_j_id = self.liste_elemente.id_Knoten_j[index_elemente]          
            
            for index_knoten in range(self.liste_knoten.Anzahl_Knoten):
                if (self.liste_knoten.id_Knoten[index_knoten] == temp_Knoten_i_id):
                    temp_Knoten_i_pos_x_in_mm = self.liste_knoten.pos_x_in_mm[index_knoten]
                    temp_Knoten_i_pos_y_in_mm = self.liste_knoten.pos_y_in_mm[index_knoten]
                    temp_Knoten_i_disp_x_in_mm = Darstellungsfaktor_Verformung * self.liste_knoten.disp_x_in_mm[index_knoten]
                    temp_Knoten_i_disp_y_in_mm = Darstellungsfaktor_Verformung * self.liste_knoten.disp_y_in_mm[index_knoten]
                if (self.liste_knoten.id_Knoten[index_knoten] == temp_Knoten_j_id):
                    temp_Knoten_j_pos_x_in_mm = self.liste_knoten.pos_x_in_mm[index_knoten]
                    temp_Knoten_j_pos_y_in_mm = self.liste_knoten.pos_y_in_mm[index_knoten]                
                    temp_Knoten_j_disp_x_in_mm = Darstellungsfaktor_Verformung * self.liste_knoten.disp_x_in_mm[index_knoten]
                    temp_Knoten_j_disp_y_in_mm = Darstellungsfaktor_Verformung * self.liste_knoten.disp_y_in_mm[index_knoten]                
            
            #Unverformter Plot
            plt.plot((temp_Knoten_i_pos_x_in_mm, temp_Knoten_j_pos_x_in_mm), (temp_Knoten_i_pos_y_in_mm, temp_Knoten_j_pos_y_in_mm), color='black', linewidth = 1)
            
            #Farbe festlegen
            Bogen = np.pi/3
            
            if abs(self.liste_elemente.sigma_linear_in_MPa[index_elemente]) >= (sigma_max_in_MPa/Sicherheitsfaktor_Auslegung):
                R=1
                G=0
                B=0
            else:
                Col_Ind = 2- 2*abs((self.liste_elemente.sigma_linear_in_MPa[index_elemente]/(sigma_max_in_MPa/Sicherheitsfaktor_Auslegung))) # Col_Ind = Color-Index        
                R= abs(np.cos(Col_Ind))
                G= abs(np.cos(Col_Ind+ 2*Bogen))
                B= abs(np.cos(Col_Ind+ 4*Bogen))
                
            plt.plot((temp_Knoten_i_pos_x_in_mm+temp_Knoten_i_disp_x_in_mm, temp_Knoten_j_pos_x_in_mm+temp_Knoten_j_disp_x_in_mm), (temp_Knoten_i_pos_y_in_mm+temp_Knoten_i_disp_y_in_mm, temp_Knoten_j_pos_y_in_mm+temp_Knoten_j_disp_y_in_mm), color=(R, G, B), linewidth = 2, linestyle = '--')

        plt.suptitle('Spannungen',fontsize = 16)   
        plt.title(' Sicherheitsfaktor Auslegung = '+str(Sicherheitsfaktor_Auslegung),fontsize = 12)      
        plt.xlabel('Koordinate x [m]')
        plt.ylabel('Koordinate y [m]')
        plt.tight_layout()
        plt.savefig('../output_bilder/' + Projektname + '_Spannung_linear.png')
        plt.close()
        
    def FEMI_K9_Plotten_ZugDruckstaebe(self):    
        for index_elemente in range(self.liste_elemente.Anzahl_Elemente):
        
            #Hinweis: Nummerierung fängt bei 1 an, Array fängt aber bei 0 an            
            temp_Knoten_i_id = self.liste_elemente.id_Knoten_i[index_elemente]            
            temp_Knoten_j_id = self.liste_elemente.id_Knoten_j[index_elemente]          
            
            for index_knoten in range(self.liste_knoten.Anzahl_Knoten):
                if (self.liste_knoten.id_Knoten[index_knoten] == temp_Knoten_i_id):
                    temp_Knoten_i_pos_x_in_mm = self.liste_knoten.pos_x_in_mm[index_knoten]
                    temp_Knoten_i_pos_y_in_mm = self.liste_knoten.pos_y_in_mm[index_knoten]
                    temp_Knoten_i_disp_x_in_mm = Darstellungsfaktor_Verformung * self.liste_knoten.disp_x_in_mm[index_knoten]
                    temp_Knoten_i_disp_y_in_mm = Darstellungsfaktor_Verformung * self.liste_knoten.disp_y_in_mm[index_knoten]
                if (self.liste_knoten.id_Knoten[index_knoten] == temp_Knoten_j_id):
                    temp_Knoten_j_pos_x_in_mm = self.liste_knoten.pos_x_in_mm[index_knoten]
                    temp_Knoten_j_pos_y_in_mm = self.liste_knoten.pos_y_in_mm[index_knoten]                
                    temp_Knoten_j_disp_x_in_mm = Darstellungsfaktor_Verformung * self.liste_knoten.disp_x_in_mm[index_knoten]
                    temp_Knoten_j_disp_y_in_mm = Darstellungsfaktor_Verformung * self.liste_knoten.disp_y_in_mm[index_knoten]                
            
            #Farbe festlegen
            Bogen = np.pi/3
            
            if self.liste_elemente.sigma_linear_in_MPa[index_elemente] >= 1e-1:
                #Zugstab
                R=0
                G=1
                B=0            
            elif self.liste_elemente.sigma_linear_in_MPa[index_elemente] <= -1e-1:
                #Druckstab
                R=1
                G=0
                B=0
            else:
                #Nullstab
                R=0
                G=0
                B=1
                            
            #Unverformter Plot
            plt.plot((temp_Knoten_i_pos_x_in_mm, temp_Knoten_j_pos_x_in_mm), (temp_Knoten_i_pos_y_in_mm, temp_Knoten_j_pos_y_in_mm), color=(R, G, B), linewidth = 1)                        
                            
        plt.suptitle('Identifikation Stäbe',fontsize = 16)   
        plt.title('grün = Zugstab, rot = Druckstab, blau = Nullstab',fontsize = 12)      
        plt.xlabel('Koordinate x [m]')
        plt.ylabel('Koordinate y [m]')
        plt.tight_layout()
        plt.savefig('../output_bilder/' + Projektname + '_Identifikation_Belastungsrichtung_Staebe.png')
        plt.close()
    
    def FEMI_K9_Schreiben_Elementinformationen_linear(self):  
        df_elemente = pd.DataFrame()

        df_elemente['Element'] = self.liste_elemente.id_Element
        df_elemente['Knoten1'] = self.liste_elemente.id_Knoten_i        
        df_elemente['Knoten2'] = self.liste_elemente.id_Knoten_j
        df_elemente['Stabkraft'] = self.liste_elemente.F_res_in_N        
        df_elemente['Knoten1 Kraft x'] = self.liste_elemente.F_i_x_in_N
        df_elemente['Knoten1 Kraft y'] = self.liste_elemente.F_i_y_in_N
        df_elemente['Knoten2 Kraft x'] = self.liste_elemente.F_j_x_in_N
        df_elemente['Knoten2 Kraft y'] = self.liste_elemente.F_j_y_in_N                
        df_elemente['Spannung linear'] = self.liste_elemente.sigma_linear_in_MPa        
        df_elemente['Verzerrung'] = self.liste_elemente.epsilon        

        self.tabelle_elemente = df_elemente.style.hide(axis='index').format({'Knoten1 Kraft x': '{:.1f}','Knoten1 Kraft y': '{:.1f}','Knoten2 Kraft x': '{:.1f}','Knoten2 Kraft y': '{:.1f}','Stabkraft': '{:.1f}','Spannung linear': '{:.3f}','Spannung nichtlinear': '{:.3f}'})
                        
        df_elemente.to_csv('../output_csv/Spannung_Elemente.csv')  
            
    def FEMII_Plotten_Konvergenz(self, datenlogger):    
        plt.semilogy(datenlogger.Iterationsnummer, datenlogger.Kraftresiduum)
        plt.suptitle('Konvergenz des Kraftresiduums',fontsize = 16)   
        plt.title('Simulationsdauer: '+str(round(self.Zeit_Dauer_Simulation_nichtlinear,3))+' s',fontsize = 12)      
        plt.xlabel('Iterationsschritte')
        plt.ylabel('Kraftresiduum')
        plt.tight_layout()
        plt.savefig('../output_bilder/' + Projektname + '_Konvergenz.png')
        plt.close()        
        
    def FEMII_Schreiben_Elementinformationen_nichtlinear(self):  
        df_elemente = pd.DataFrame()

        df_elemente['Element'] = self.liste_elemente.id_Element
        df_elemente['Knoten1'] = self.liste_elemente.id_Knoten_i        
        df_elemente['Knoten2'] = self.liste_elemente.id_Knoten_j
        df_elemente['Stabkraft'] = self.liste_elemente.F_res_in_N        
        df_elemente['Knoten1 Kraft x'] = self.liste_elemente.F_i_x_in_N
        df_elemente['Knoten1 Kraft y'] = self.liste_elemente.F_i_y_in_N
        df_elemente['Knoten2 Kraft x'] = self.liste_elemente.F_j_x_in_N
        df_elemente['Knoten2 Kraft y'] = self.liste_elemente.F_j_y_in_N
        df_elemente['Spannung nichtlinear'] = self.liste_elemente.sigma_nichtlinear_in_MPa
        df_elemente['Verzerrung nichtlinear'] = self.liste_elemente.epsilon        
        
        self.tabelle_elemente = df_elemente.style.hide(axis='index').format({'Knoten1 Kraft x': '{:.1f}','Knoten1 Kraft y': '{:.1f}','Knoten2 Kraft x': '{:.1f}','Knoten2 Kraft y': '{:.1f}','Stabkraft': '{:.1f}','Spannung linear': '{:.3f}','Spannung nichtlinear': '{:.3f}'})
                        
        df_elemente.to_csv('../output_csv/Spannung_Elemente.csv')                  
     
    def FEMII_Plotten_Spannungen_nichtlinear(self):    
        for index_elemente in range(self.liste_elemente.Anzahl_Elemente):
        
            #Hinweis: Nummerierung fängt bei 1 an, Array fängt aber bei 0 an            
            temp_Knoten_i_id = self.liste_elemente.id_Knoten_i[index_elemente]            
            temp_Knoten_j_id = self.liste_elemente.id_Knoten_j[index_elemente]          
            
            for index_knoten in range(self.liste_knoten.Anzahl_Knoten):
                if (self.liste_knoten.id_Knoten[index_knoten] == temp_Knoten_i_id):
                    temp_Knoten_i_pos_x_in_mm = self.liste_knoten.pos_x_in_mm[index_knoten]
                    temp_Knoten_i_pos_y_in_mm = self.liste_knoten.pos_y_in_mm[index_knoten]
                    temp_Knoten_i_disp_x_in_mm = Darstellungsfaktor_Verformung * self.liste_knoten.disp_x_in_mm[index_knoten]
                    temp_Knoten_i_disp_y_in_mm = Darstellungsfaktor_Verformung * self.liste_knoten.disp_y_in_mm[index_knoten]
                if (self.liste_knoten.id_Knoten[index_knoten] == temp_Knoten_j_id):
                    temp_Knoten_j_pos_x_in_mm = self.liste_knoten.pos_x_in_mm[index_knoten]
                    temp_Knoten_j_pos_y_in_mm = self.liste_knoten.pos_y_in_mm[index_knoten]                
                    temp_Knoten_j_disp_x_in_mm = Darstellungsfaktor_Verformung * self.liste_knoten.disp_x_in_mm[index_knoten]
                    temp_Knoten_j_disp_y_in_mm = Darstellungsfaktor_Verformung * self.liste_knoten.disp_y_in_mm[index_knoten]                
            
            #Unverformter Plot
            plt.plot((temp_Knoten_i_pos_x_in_mm, temp_Knoten_j_pos_x_in_mm), (temp_Knoten_i_pos_y_in_mm, temp_Knoten_j_pos_y_in_mm), color='black', linewidth = 1)
            
            #Farbe festlegen
            Bogen = np.pi/3
            
            if abs(self.liste_elemente.sigma_nichtlinear_in_MPa[index_elemente]) >= (sigma_max_in_MPa/Sicherheitsfaktor_Auslegung):
                R=1
                G=0
                B=0
            else:
                Col_Ind = 2- 2*abs((self.liste_elemente.sigma_nichtlinear_in_MPa[index_elemente]/(sigma_max_in_MPa/Sicherheitsfaktor_Auslegung))) # Col_Ind = Color-Index        
                R= abs(np.cos(Col_Ind))
                G= abs(np.cos(Col_Ind+ 2*Bogen))
                B= abs(np.cos(Col_Ind+ 4*Bogen))
                
            plt.plot((temp_Knoten_i_pos_x_in_mm+temp_Knoten_i_disp_x_in_mm, temp_Knoten_j_pos_x_in_mm+temp_Knoten_j_disp_x_in_mm), (temp_Knoten_i_pos_y_in_mm+temp_Knoten_i_disp_y_in_mm, temp_Knoten_j_pos_y_in_mm+temp_Knoten_j_disp_y_in_mm), color=(R, G, B), linewidth = 2, linestyle = '--')

        plt.suptitle('Spannungen - nichtlinear gerechnet',fontsize = 16)   
        plt.title(' Sicherheitsfaktor Auslegung = '+str(Sicherheitsfaktor_Auslegung),fontsize = 12)      
        plt.xlabel('Koordinate x [m]')
        plt.ylabel('Koordinate y [m]')
        plt.tight_layout()
        plt.savefig('../output_bilder/' + Projektname + '_Spannung_nichtlinear.png')
        plt.close()
        
    def FEMII_Plotten_Verformung_nichtlinear(self):
        for index_elemente in range(self.liste_elemente.Anzahl_Elemente):
    
            #Hinweis: Nummerierung fängt bei 1 an, Array fängt aber bei 0 an            
            temp_Knoten_i_id = self.liste_elemente.id_Knoten_i[index_elemente]            
            temp_Knoten_j_id = self.liste_elemente.id_Knoten_j[index_elemente]          
            
            for index_knoten in range(self.liste_knoten.Anzahl_Knoten):
                if (self.liste_knoten.id_Knoten[index_knoten] == temp_Knoten_i_id):
                    temp_Knoten_i_pos_x_in_mm = self.liste_knoten.pos_x_in_mm[index_knoten]
                    temp_Knoten_i_pos_y_in_mm = self.liste_knoten.pos_y_in_mm[index_knoten]
                    temp_Knoten_i_disp_x_in_mm = Darstellungsfaktor_Verformung * self.liste_knoten.disp_x_in_mm[index_knoten]
                    temp_Knoten_i_disp_y_in_mm = Darstellungsfaktor_Verformung * self.liste_knoten.disp_y_in_mm[index_knoten]
                if (self.liste_knoten.id_Knoten[index_knoten] == temp_Knoten_j_id):
                    temp_Knoten_j_pos_x_in_mm = self.liste_knoten.pos_x_in_mm[index_knoten]
                    temp_Knoten_j_pos_y_in_mm = self.liste_knoten.pos_y_in_mm[index_knoten]                
                    temp_Knoten_j_disp_x_in_mm = Darstellungsfaktor_Verformung * self.liste_knoten.disp_x_in_mm[index_knoten]
                    temp_Knoten_j_disp_y_in_mm = Darstellungsfaktor_Verformung * self.liste_knoten.disp_y_in_mm[index_knoten]                
            
            temp_max_disp_x = abs(max(self.liste_knoten.disp_x_in_mm))
            temp_min_disp_x = abs(min(self.liste_knoten.disp_x_in_mm))
            
            if (temp_max_disp_x > temp_min_disp_x):
                temp_max_abs_disp_x = temp_max_disp_x
            else:
                temp_max_abs_disp_x = temp_min_disp_x
                
            temp_max_disp_y = abs(max(self.liste_knoten.disp_y_in_mm))
            temp_min_disp_y = abs(min(self.liste_knoten.disp_y_in_mm))
            
            if (temp_max_disp_y > temp_min_disp_y):
                temp_max_abs_disp_y = temp_max_disp_y
            else:
                temp_max_abs_disp_y = temp_min_disp_y
            
            temp_max_abs_disp_res = np.sqrt(pow(temp_max_abs_disp_x,2)+pow(temp_max_abs_disp_y,2))
            #Unverformter Plot
            plt.plot((temp_Knoten_i_pos_x_in_mm, temp_Knoten_j_pos_x_in_mm), (temp_Knoten_i_pos_y_in_mm, temp_Knoten_j_pos_y_in_mm), color='black', linewidth = 1)
                        
            plt.plot((temp_Knoten_i_pos_x_in_mm+temp_Knoten_i_disp_x_in_mm, temp_Knoten_j_pos_x_in_mm+temp_Knoten_j_disp_x_in_mm), (temp_Knoten_i_pos_y_in_mm+temp_Knoten_i_disp_y_in_mm, temp_Knoten_j_pos_y_in_mm+temp_Knoten_j_disp_y_in_mm), color='blue', linewidth = 2, linestyle = '--')

        plt.suptitle('Verformung - nichtlinear gerechnet',fontsize = 16)   
        plt.title('Darstellungsfaktor = '+str(Darstellungsfaktor_Verformung)+ ', max. Verschiebung = '+str(round(temp_max_abs_disp_res,2))+' mm',fontsize = 12 )   
        plt.xlabel('Koordinate x [m]')
        plt.ylabel('Koordinate y [m]')  
        plt.tight_layout()        
        plt.savefig('../output_bilder/' + Projektname + '_Verformung_nichtlinear.png')
        plt.close()
              
    ###########################################################################
    #### Preprocessor-Funktionen ##############################################
    ###########################################################################
    
    def Startzeit_Simulation_bestimmen_linear(self):
        self.Zeit_Start_Simulation_linear = time.time()
        
    def Endzeit_Simulation_bestimmen_linear(self):
        self.Zeit_Ende_Simulation_linear = time.time()
        self.Zeit_Dauer_Simulation_linear = self.Zeit_Ende_Simulation_linear - self.Zeit_Start_Simulation_linear
        print("Simultionszeit lineare Simulation: "+str(self.Zeit_Dauer_Simulation_linear))
        
    def Startzeit_Simulation_bestimmen_nichtlinear(self):
        self.Zeit_Start_Simulation_nichtlinear = time.time()
        
    def Endzeit_Simulation_bestimmen_nichtlinear(self):
        self.Zeit_Ende_Simulation_nichtlinear = time.time()
        self.Zeit_Dauer_Simulation_nichtlinear = self.Zeit_Ende_Simulation_nichtlinear - self.Zeit_Start_Simulation_nichtlinear
        print("Simultionszeit nichtlineare Simulation: "+str(self.Zeit_Dauer_Simulation_nichtlinear))
        
    def FEMI_K2_Berechnung_Elementlaengen(self):
        for index_elemente in range(self.liste_elemente.Anzahl_Elemente):
                
            temp_Knoten_i_id = self.liste_elemente.id_Knoten_i[index_elemente]            
            temp_Knoten_j_id = self.liste_elemente.id_Knoten_j[index_elemente]          
            
            for index_knoten in range(self.liste_knoten.Anzahl_Knoten):
                if (self.liste_knoten.id_Knoten[index_knoten] == temp_Knoten_i_id):
                    temp_Knoten_i_x_in_mm = self.liste_knoten.pos_x_in_mm[index_knoten]
                    temp_Knoten_i_y_in_mm = self.liste_knoten.pos_y_in_mm[index_knoten]
                if (self.liste_knoten.id_Knoten[index_knoten] == temp_Knoten_j_id):
                    temp_Knoten_j_x_in_mm = self.liste_knoten.pos_x_in_mm[index_knoten]
                    temp_Knoten_j_y_in_mm = self.liste_knoten.pos_y_in_mm[index_knoten]   
                    
            self.liste_elemente.L_in_mm[index_elemente] = round(np.sqrt((temp_Knoten_i_x_in_mm-temp_Knoten_j_x_in_mm)**2 + (temp_Knoten_i_y_in_mm-temp_Knoten_j_y_in_mm)**2),2)
            
        if (Ausgaben_schreiben == True):
            print("L_in_mm:")
            print(self.liste_elemente.L_in_mm)
            print("\n")
            
    def FEMI_K3_K4_Berechnung_Formfunktion(self):        
        ### Definition der x-Werte der Stützpunkte
        xi_smp = smp.symbols("xi")
        self.xi = np.linspace(-1, 1, polynomgrad+1)
        
        if (Ausgaben_schreiben == True):
            print("xi:")
            print(str(self.xi))  
            print("\n")        

        Variable_XX = [[0.0 for _ in range(polynomgrad)] for _ in range(polynomgrad+1)]

        Variable_XX = np.array(Variable_XX)

        for z in range(polynomgrad+1):
            i = 0
            
            for s in range(polynomgrad+1):
                if s != z:
                    
                    Variable_XX[z, i] += self.xi[s]
                    
                    i += 1

        if (Ausgaben_schreiben == True):
            print("Variable_XX:")
            print(Variable_XX) 
            print("\n")
        
        ### Berechnung und Darstellung der Formfunktionen
        self.N = zeros(polynomgrad+1, 1)

        smpplt = smp.plot(title='Formfunktion', show=False)


        for index_p in range(polynomgrad+1):
            
            Variable_L = 1
            
            for term in range(polynomgrad):
                
                Variable_L = Variable_L * ((Variable_XX[index_p, term] - xi_smp) / (Variable_XX[index_p, term] - self.xi[index_p]))
                
            self.N[index_p] += Variable_L
            
            smpplt.extend(smp.plot(self.N[index_p], (xi_smp, -1, 1), show=False))

        if (Ausgaben_schreiben == True):
            smpplt.show()

        if (Ausgaben_schreiben == True):
            print("N:")
            print(self.N)
            print("\n")
            
        ### Berechnung der Ableitung der Formfunktionen
        self.B = smp.diff(self.N, xi_smp)

        if (Ausgaben_schreiben == True):
            print("B:")
            print(self.B)
            print("\n")
        
        self.B_transponiert = Matrix(self.B).transpose()

        if (Ausgaben_schreiben == True):
            print("B_transponiert:")
            print(self.B_transponiert)
            print("\n")
                
            
    def FEMI_K3_K4_Berechnung_Normierte_Elementsteifigkeitsmatrix(self):                        
        ## Erstellung der normierten Elementsteifigkeitsmatrix eines Stabelementes
        xi_smp = smp.symbols("xi")
        
        K = self.B * self.B_transponiert
        
        if (Ausgaben_schreiben == True):
            print("K:")
            print(K)
            print("\n")
        
        I_b = smp.integrate(K, (xi_smp, -1 ,1))
        
        if (Ausgaben_schreiben == True):
            print("I_b:")
            print(I_b)
            print("\n")
        
        KM = zeros((polynomgrad+1)*2)

        for zeile in range(polynomgrad+1):
            
            for spalte in range(polynomgrad+1):
                
                KM[zeile*2, spalte*2] += I_b[zeile, spalte]

        KM_array = np.array(KM).astype(np.float64)

        self.KE_normiert = KM_array
        
        if (Ausgaben_schreiben == True):
            print("KE_normiert")
            print(self.KE_normiert)
            print("\n")

    def FEMI_K5_Berechnung_Absolute_Elementsteifigkeitsmatrizen(self):
        for index_elemente in range(self.liste_elemente.Anzahl_Elemente):
                
            temp_Element_id = self.liste_elemente.id_Element[index_elemente]            
            
            temp_Knoten_i_id = self.liste_elemente.id_Knoten_i[index_elemente]            
            temp_Knoten_j_id = self.liste_elemente.id_Knoten_j[index_elemente]          
            
            for index_knoten in range(self.liste_knoten.Anzahl_Knoten):
                if (self.liste_knoten.id_Knoten[index_knoten] == temp_Knoten_i_id):
                    temp_Knoten_i_pos_x_in_mm = self.liste_knoten.pos_x_in_mm[index_knoten]
                    temp_Knoten_i_pos_y_in_mm = self.liste_knoten.pos_y_in_mm[index_knoten]            
                if (self.liste_knoten.id_Knoten[index_knoten] == temp_Knoten_j_id):
                    temp_Knoten_j_pos_x_in_mm = self.liste_knoten.pos_x_in_mm[index_knoten]
                    temp_Knoten_j_pos_y_in_mm = self.liste_knoten.pos_y_in_mm[index_knoten] 
                            
            
            # print('relative Laenge:', L_rel)
            
            C = (temp_Knoten_j_pos_x_in_mm-temp_Knoten_i_pos_x_in_mm)/self.liste_elemente.L_in_mm[index_elemente] # Cosinus-Anteil, x
            S = (temp_Knoten_j_pos_y_in_mm-temp_Knoten_i_pos_y_in_mm)/self.liste_elemente.L_in_mm[index_elemente] # Sinus-Anteil, y
                        
            self.liste_elemente.C[index_elemente] = C
            self.liste_elemente.S[index_elemente] = S                       
            
            K_tilde = self.KE_normiert            
            
            T = np.array([[C, S, 0, 0], [-S, C, 0, 0], [0, 0, C, S], [0, 0, -S, C]])
            
            T_transponiert = T.transpose()
            
            K_transponiert = np.matmul(K_tilde, T)
                        
            temp_E_in_MPa = self.liste_elemente.E_in_MPa[index_elemente]                        
            
            self.liste_elemente.KE[index_elemente] = np.matmul(T_transponiert, K_transponiert) * 2 * temp_E_in_MPa * self.liste_elemente.A_in_sqmm[index_elemente] / (self.liste_elemente.L_in_mm[index_elemente]) # Elementsteifigkeitsmatrix    
            
            if (Ausgaben_schreiben == True):
                print('Elementsteifigkeitsmatrix Element '+str(temp_Element_id)+':\n', self.liste_elemente.KE[index_elemente])       
                print('\n') 

    def FEMI_K6_Berechnung_Gesamtsteifigkeitsmatrix(self):
        ## Erstellen einer leeren Gesamtsteifigkeitsmatrix
        self.K = [[0.0 for _ in range(self.liste_knoten.Anzahl_Knoten*2)] for _ in range(self.liste_knoten.Anzahl_Knoten*2)]
        self.K = np.array(self.K)
        
        ## Einarbeitung der Elementsteifigkeitsmatritzen in die Gesamtsteifigkeitsmatrix        
        for index_elemente in range(self.liste_elemente.Anzahl_Elemente):
        
            temp_Knoten_i_id = self.liste_elemente.id_Knoten_i[index_elemente]            
            temp_Knoten_j_id = self.liste_elemente.id_Knoten_j[index_elemente]          
                    
            # Viertelung der globalen Elementsteifigkeitsmatrix
            # oben = x-Koordinate, unten = y-Koordinate
            # links = erster Knoten, rechts = zweiter Knoten
            
            K_ol = self.liste_elemente.KE[index_elemente][0:2, 0:2] # oben/links
            K_or = self.liste_elemente.KE[index_elemente][0:2, 2:4] # oben/rechts
            K_ul = self.liste_elemente.KE[index_elemente][2:4, 0:2] # unten/links
            K_ur = self.liste_elemente.KE[index_elemente][2:4, 2:4] # unten/rechts
            
            
            self.K[2*temp_Knoten_i_id-2: 2*temp_Knoten_i_id, 2*temp_Knoten_i_id-2: 2*temp_Knoten_i_id] += K_ol
            self.K[2*temp_Knoten_i_id-2: 2*temp_Knoten_i_id, 2*temp_Knoten_j_id-2: 2*temp_Knoten_j_id] += K_or
            self.K[2*temp_Knoten_j_id-2: 2*temp_Knoten_j_id, 2*temp_Knoten_i_id-2: 2*temp_Knoten_i_id] += K_ul
            self.K[2*temp_Knoten_j_id-2: 2*temp_Knoten_j_id, 2*temp_Knoten_j_id-2: 2*temp_Knoten_j_id] += K_ur
        
        if (Ausgaben_schreiben == True):        
            print("K:")
            print(self.K)
            print("\n")
        
        ### Ausgeben der globalen Gesamtsteifigkeitsmatrix
        
        Kdf = pd.DataFrame()

        # spaltenweises Einschreiben der Steifigkeitsmatrix in ein DataFrame

        for c in range(len(self.K)):
            
            K_col = pd.Series(self.K[:, c])
            
            cc = c + 1
            
            Kdf[f'{cc}. Spalte'] = pd.DataFrame(K_col)


        Kdf_round = Kdf.round(3)
            
        Kdf_round.style.hide(axis='index')
    
    def FEMI_K7_Berechnung_Lagerungen_in_Gesamtsteifigkeitsmatrix(self):
        self.K_Lager = self.K
        for index_knoten in range(self.liste_knoten.Anzahl_Knoten):
            if (self.liste_knoten.Lager_x[index_knoten] == 1): 
                self.K_Lager[index_knoten*2, :] = 0
                self.K_Lager[:, index_knoten*2] = 0
                self.K_Lager[index_knoten*2, index_knoten*2] = 1
            if (self.liste_knoten.Lager_y[index_knoten] == 1): 
                self.K_Lager[index_knoten*2+1, :] = 0
                self.K_Lager[:, index_knoten*2+1] = 0
                self.K_Lager[index_knoten*2+1, index_knoten*2+1] = 1
        
        if (Ausgaben_schreiben == True):
            print("K_Lager:")
            print(self.K_Lager)
            print("\n")
            
    def FEMI_K7_Berechnung_Kraftvektor(self):
        self.F = [] # leerer Vektor in den geschrieben wird

        for index_knoten in range(self.liste_knoten.Anzahl_Knoten):
                       
            self.F.append(self.liste_knoten.F_x_stern_in_N[index_knoten]) # Kräfte in x-Richtung
                                     
            self.F.append(self.liste_knoten.F_y_stern_in_N[index_knoten]) # Kräfte in y-Richtung

        self.F = np.array(self.F)

        if (Ausgaben_schreiben == True):
            print("F:")
            print(self.F)
            print("\n")
            
            
    def FEMI_K8_Berechnung_Inverse_Steifigkeitsmatrix(self):                    
        self.K_Lager_inv = np.linalg.inv(self.K_Lager) # Inverse der Steifigkeitsmatrix
        
    def FEMI_K8_Berechnung_Knotenverschiebungsmatrix(self):                            
        self.u = np.matmul(self.K_Lager_inv, self.F) # Verschiebungsvektor u

        if (Ausgaben_schreiben == True):
            print("u:")
            print(self.u)
            print("\n")
        
        for index_knoten in range(self.liste_knoten.Anzahl_Knoten):
            self.liste_knoten.disp_x_in_mm[index_knoten] = self.u[index_knoten*2]
            self.liste_knoten.disp_y_in_mm[index_knoten] = self.u[index_knoten*2+1]    

        if (Ausgaben_schreiben == True):                         
            print("disp_x_in_mm:")
            print(self.liste_knoten.disp_x_in_mm)
            print("\n")
            
            print("disp_y_in_mm:")
            print(self.liste_knoten.disp_y_in_mm)
            print("\n")
    
    def FEMI_K9_Berechnung_Verzerrungen(self):
        for index_elemente in range(self.liste_elemente.Anzahl_Elemente):
            
            #Hinweis: Nummerierung fängt bei 1 an, Array fängt aber bei 0 an            
            temp_Knoten_i_id = self.liste_elemente.id_Knoten_i[index_elemente]            
            temp_Knoten_j_id = self.liste_elemente.id_Knoten_j[index_elemente]          
            
            for index_knoten in range(self.liste_knoten.Anzahl_Knoten):
                if (self.liste_knoten.id_Knoten[index_knoten] == temp_Knoten_i_id):
                    temp_Knoten_i_x_in_mm = self.liste_knoten.pos_x_in_mm[index_knoten]
                    temp_Knoten_i_y_in_mm = self.liste_knoten.pos_y_in_mm[index_knoten]
                if (self.liste_knoten.id_Knoten[index_knoten] == temp_Knoten_j_id):
                    temp_Knoten_j_x_in_mm = self.liste_knoten.pos_x_in_mm[index_knoten]
                    temp_Knoten_j_y_in_mm = self.liste_knoten.pos_y_in_mm[index_knoten]                    
              
            
            C = (temp_Knoten_j_x_in_mm-temp_Knoten_i_x_in_mm)/self.liste_elemente.L_in_mm[index_elemente] # Cosinus-Anteil, x
            S = (temp_Knoten_j_y_in_mm-temp_Knoten_i_y_in_mm)/self.liste_elemente.L_in_mm[index_elemente] # Sinus-Anteil, y
            
            BB = np.array([-C, -S,  C, S]) / self.liste_elemente.L_in_mm[index_elemente]
            
            scatter = [2*temp_Knoten_i_id-2, 2*temp_Knoten_i_id-1, 2*temp_Knoten_j_id-2, 2*temp_Knoten_j_id-1] # scatter [engl.: streuen, zerstreuen]
            
            D_sc = []

            for s in scatter:

                D_sc.append(self.u[s])
            
            DD = np.array(D_sc).transpose()
            
            self.liste_elemente.epsilon[index_elemente] = float(BB @ DD)    
        
        if (Ausgaben_schreiben == True):                 
            print("epsilon:")
            print(self.liste_elemente.epsilon)
            print("\n")
    
    def FEMI_K9_Berechnung_Spannungen_linear(self):
        self.liste_elemente.sigma_linear_in_MPa = self.liste_elemente.E_in_MPa * np.array(self.liste_elemente.epsilon)
                
        if (Physikalische_Nichtlinearitaet_beruecksichtigen == True):
            for index_elemente in range(self.liste_elemente.Anzahl_Elemente):
                self.liste_elemente.sigma_nichtlinear_in_MPa[index_elemente] = self.bibliothek_materialien.Material[self.liste_elemente.Material_id[index_elemente]].f_epsilon_sigma(self.liste_elemente.epsilon[index_elemente])    
                
            
        if (Ausgaben_schreiben == True):
            print("sigma_linear_in_MPa:")
            print(self.liste_elemente.sigma_linear_in_MPa)
            print("\n")
            
    def FEMI_K9_Berechnung_innere_Reaktionskraefte(self):
        #Erst einmal alle Reaktiosnkräfte zu Null setzen, da danach aufintegriert wird
        for index_knoten in range(self.liste_knoten.Anzahl_Knoten):
            self.liste_knoten.F_x_reactio_in_N[index_knoten] = 0
            self.liste_knoten.F_y_reactio_in_N[index_knoten] = 0
            self.liste_knoten.F_x_reactio_in_N[index_knoten] = 0
            self.liste_knoten.F_y_reactio_in_N[index_knoten] = 0
            
        for index_elemente in range(self.liste_elemente.Anzahl_Elemente):
            temp_Knoten_i_id = self.liste_elemente.id_Knoten_i[index_elemente]            
            temp_Knoten_j_id = self.liste_elemente.id_Knoten_j[index_elemente]          
            
            self.liste_elemente.F_res_in_N[index_elemente] = self.liste_elemente.sigma_linear_in_MPa[index_elemente] * self.liste_elemente.A_in_sqmm[index_elemente]
            
            self.liste_elemente.F_i_x_in_N[index_elemente] = self.liste_elemente.F_res_in_N[index_elemente] * self.liste_elemente.C[index_elemente]
            self.liste_elemente.F_i_y_in_N[index_elemente] = self.liste_elemente.F_res_in_N[index_elemente] * self.liste_elemente.S[index_elemente]
            self.liste_elemente.F_j_x_in_N[index_elemente] = - self.liste_elemente.F_i_x_in_N[index_elemente]
            self.liste_elemente.F_j_y_in_N[index_elemente] = - self.liste_elemente.F_i_y_in_N[index_elemente]
            
            self.liste_knoten.F_x_reactio_in_N[temp_Knoten_i_id-1] -= self.liste_elemente.F_i_x_in_N[index_elemente]
            self.liste_knoten.F_y_reactio_in_N[temp_Knoten_i_id-1] -= self.liste_elemente.F_i_y_in_N[index_elemente]
            self.liste_knoten.F_x_reactio_in_N[temp_Knoten_j_id-1] -= self.liste_elemente.F_j_x_in_N[index_elemente]
            self.liste_knoten.F_y_reactio_in_N[temp_Knoten_j_id-1] -= self.liste_elemente.F_j_y_in_N[index_elemente]            
        
        if (Ausgaben_schreiben == True):        
            print("liste_knoten.F_x_reactio_in_N:")
            print(self.liste_knoten.F_x_reactio_in_N)
            print("\n")
            
            print("liste_knoten.F_y_reactio_in_N:")
            print(self.liste_knoten.F_y_reactio_in_N)
            print("\n")
    
    def FEMI_K9_Berechnung_innere_Reaktionskraefte_Lagerkraefte_loeschen(self):
        for index_knoten in range(self.liste_knoten.Anzahl_Knoten):
            if (self.liste_knoten.Lager_x[index_knoten] == 1):
                self.liste_knoten.F_x_reactio_in_N[index_knoten] = 0
            if (self.liste_knoten.Lager_y[index_knoten] == 1):
                self.liste_knoten.F_y_reactio_in_N[index_knoten] = 0
        
        if (Ausgaben_schreiben == True):                
            print("liste_knoten.F_x_reactio_in_N:")
            print(self.liste_knoten.F_x_reactio_in_N)
            print("\n")
            
            print("liste_knoten.F_y_reactio_in_N:")
            print(self.liste_knoten.F_y_reactio_in_N)
            print("\n")
            
    def FEMII_Berechnung_Spannungen_nichtlinear(self):               
        if (Physikalische_Nichtlinearitaet_beruecksichtigen == True):
            for index_elemente in range(self.liste_elemente.Anzahl_Elemente):
                self.liste_elemente.sigma_nichtlinear_in_MPa[index_elemente] = self.bibliothek_materialien.Material[self.liste_elemente.Material_id[index_elemente]].f_epsilon_sigma(self.liste_elemente.epsilon[index_elemente])    
            
        if (Ausgaben_schreiben == True):
            print("sigma_nichtlinear_in_MPa:")
            print(self.liste_elemente.sigma_nichtlinear_in_MPa)
            print("\n")
      
    def FEMII_Berechnung_innere_Reaktionskraefte_nichtlinear(self):
        #Erst einmal alle Reaktiosnkräfte zu Null setzen, da danach aufintegriert wird
        for index_knoten in range(self.liste_knoten.Anzahl_Knoten):
            self.liste_knoten.F_x_reactio_in_N[index_knoten] = 0
            self.liste_knoten.F_y_reactio_in_N[index_knoten] = 0
            self.liste_knoten.F_x_reactio_in_N[index_knoten] = 0
            self.liste_knoten.F_y_reactio_in_N[index_knoten] = 0
            
        for index_elemente in range(self.liste_elemente.Anzahl_Elemente):
            temp_Knoten_i_id = self.liste_elemente.id_Knoten_i[index_elemente]            
            temp_Knoten_j_id = self.liste_elemente.id_Knoten_j[index_elemente]          
                        
            if (Physikalische_Nichtlinearitaet_beruecksichtigen == True):
                self.liste_elemente.F_res_in_N[index_elemente] = self.liste_elemente.sigma_nichtlinear_in_MPa[index_elemente] * self.liste_elemente.A_in_sqmm[index_elemente]
            
            self.liste_elemente.F_i_x_in_N[index_elemente] = self.liste_elemente.F_res_in_N[index_elemente] * self.liste_elemente.C[index_elemente]
            self.liste_elemente.F_i_y_in_N[index_elemente] = self.liste_elemente.F_res_in_N[index_elemente] * self.liste_elemente.S[index_elemente]
            self.liste_elemente.F_j_x_in_N[index_elemente] = - self.liste_elemente.F_i_x_in_N[index_elemente]
            self.liste_elemente.F_j_y_in_N[index_elemente] = - self.liste_elemente.F_i_y_in_N[index_elemente]
            
            self.liste_knoten.F_x_reactio_in_N[temp_Knoten_i_id-1] -= self.liste_elemente.F_i_x_in_N[index_elemente]
            self.liste_knoten.F_y_reactio_in_N[temp_Knoten_i_id-1] -= self.liste_elemente.F_i_y_in_N[index_elemente]
            self.liste_knoten.F_x_reactio_in_N[temp_Knoten_j_id-1] -= self.liste_elemente.F_j_x_in_N[index_elemente]
            self.liste_knoten.F_y_reactio_in_N[temp_Knoten_j_id-1] -= self.liste_elemente.F_j_y_in_N[index_elemente]            
        
        if (Ausgaben_schreiben == True):        
            print("liste_knoten.F_x_reactio_in_N:")
            print(self.liste_knoten.F_x_reactio_in_N)
            print("\n")
            
            print("liste_knoten.F_y_reactio_in_N:")
            print(self.liste_knoten.F_y_reactio_in_N)
            print("\n")
            
    def FEMII_Berechnung_Differenzkraftvektor(self):                                   
        for index_knoten in range(self.liste_knoten.Anzahl_Knoten):
            self.liste_knoten.delta_F_x_in_N[index_knoten] = self.liste_knoten.F_x_stern_in_N[index_knoten] - self.liste_knoten.F_x_reactio_in_N[index_knoten]
            self.liste_knoten.delta_F_y_in_N[index_knoten] = self.liste_knoten.F_y_stern_in_N[index_knoten] - self.liste_knoten.F_y_reactio_in_N[index_knoten]

        self.delta_F = [] # leerer Vektor in den geschrieben wird

        for index_knoten in range(self.liste_knoten.Anzahl_Knoten):                       
            self.delta_F.append(self.liste_knoten.delta_F_x_in_N[index_knoten]) # Kräfte in x-Richtung                                     
            self.delta_F.append(self.liste_knoten.delta_F_y_in_N[index_knoten]) # Kräfte in y-Richtung

        self.delta_F = np.array(self.delta_F)
        
        if (Ausgaben_schreiben == True):                     
            print("delta_F_x_in_N:")
            print(self.liste_knoten.delta_F_x_in_N)
            print("\n")   
            
            print("delta_F_y_in_N:")
            print(self.liste_knoten.delta_F_y_in_N)
            print("\n")   
            
        if (Ausgaben_schreiben == True):
            print("delta_F:")
            print(self.delta_F)
            print("\n")
    
    def FEMII_Berechnung_Differenzverschiebungsmatrix(self):                            
        self.delta_u = np.matmul(self.K_Lager_inv, self.delta_F) # Verschiebungsvektor u

        if (Ausgaben_schreiben == True):
            print("delta_u:")
            print(self.delta_u)
            print("\n")
        
        for index_knoten in range(self.liste_knoten.Anzahl_Knoten):
            self.liste_knoten.delta_disp_x_in_mm[index_knoten] = self.delta_u[index_knoten*2]
            self.liste_knoten.delta_disp_y_in_mm[index_knoten] = self.delta_u[index_knoten*2+1]    

        if (Ausgaben_schreiben == True):            
            print("delta_disp_x_in_mm:")
            print(self.liste_knoten.delta_disp_x_in_mm)
            print("\n")
            
            print("delta_disp_y_in_mm:")
            print(self.liste_knoten.delta_disp_y_in_mm)
            print("\n")
    
    def FEMII_Berechnung_Addition_Differenzverschiebungsmatrix_in_Verschiebungsmatrix(self):                                                   
        #Update von u zur Berechnung der Verzerrungen        
        for index_knoten in range(self.liste_knoten.Anzahl_Knoten):    
            self.u[index_knoten*2] += (self.liste_knoten.delta_disp_x_in_mm[index_knoten])
            self.u[index_knoten*2+1] += (self.liste_knoten.delta_disp_y_in_mm[index_knoten])
        
        for index_knoten in range(self.liste_knoten.Anzahl_Knoten):
            self.liste_knoten.disp_x_in_mm[index_knoten] = self.u[index_knoten*2]
            self.liste_knoten.disp_y_in_mm[index_knoten] = self.u[index_knoten*2+1]            
            
        if (Ausgaben_schreiben == True): 
            print("u:")
            print(self.u)
            print("\n")
            
            print("disp_x_in_mm:")
            print(self.liste_knoten.disp_x_in_mm)
            print("\n")
            
            print("disp_y_in_mm:")
            print(self.liste_knoten.disp_y_in_mm)
            print("\n")
            
    def FEMII_Berechnung_Kraftresiduum(self):                   
        self.Residuum_F_K = 0
        
        for index_knoten in range(self.liste_knoten.Anzahl_Knoten):
            self.Residuum_F_K += abs(self.liste_knoten.delta_F_x_in_N[index_knoten])
            self.Residuum_F_K += abs(self.liste_knoten.delta_F_y_in_N[index_knoten])

        if (Ausgaben_schreiben == True):                     
            print("Residuum_F_K:")
            print(self.Residuum_F_K)
            print("\n")        
            
    def FEMII_Berechnung_absolute_tangentiale_Elementsteifigkeitsmatrizen(self):
        for index_elemente in range(self.liste_elemente.Anzahl_Elemente):
                
            temp_Element_id = self.liste_elemente.id_Element[index_elemente]            
            
            temp_Knoten_i_id = self.liste_elemente.id_Knoten_i[index_elemente]            
            temp_Knoten_j_id = self.liste_elemente.id_Knoten_j[index_elemente]          
            
            for index_knoten in range(self.liste_knoten.Anzahl_Knoten):
                if (self.liste_knoten.id_Knoten[index_knoten] == temp_Knoten_i_id):
                    temp_Knoten_i_pos_x_in_mm = self.liste_knoten.pos_x_in_mm[index_knoten]
                    temp_Knoten_i_pos_y_in_mm = self.liste_knoten.pos_y_in_mm[index_knoten]            
                if (self.liste_knoten.id_Knoten[index_knoten] == temp_Knoten_j_id):
                    temp_Knoten_j_pos_x_in_mm = self.liste_knoten.pos_x_in_mm[index_knoten]
                    temp_Knoten_j_pos_y_in_mm = self.liste_knoten.pos_y_in_mm[index_knoten] 
                            
            
            # print('relative Laenge:', L_rel)
            
            C = (temp_Knoten_j_pos_x_in_mm-temp_Knoten_i_pos_x_in_mm)/self.liste_elemente.L_in_mm[index_elemente] # Cosinus-Anteil, x
            S = (temp_Knoten_j_pos_y_in_mm-temp_Knoten_i_pos_y_in_mm)/self.liste_elemente.L_in_mm[index_elemente] # Sinus-Anteil, y
                        
            self.liste_elemente.C[index_elemente] = C
            self.liste_elemente.S[index_elemente] = S                       
            
            K_tilde = self.KE_normiert            
            
            T = np.array([[C, S, 0, 0], [-S, C, 0, 0], [0, 0, C, S], [0, 0, -S, C]])
            
            T_transponiert = T.transpose()
            
            K_transponiert = np.matmul(K_tilde, T)
                                                
            temp_E_in_MPa = self.bibliothek_materialien.Material[self.liste_elemente.Material_id[index_elemente]].f_E(self.liste_elemente.epsilon[index_elemente])
            
            self.liste_elemente.KE[index_elemente] = np.matmul(T_transponiert, K_transponiert) * 2 * self.liste_elemente.A_in_sqmm[index_elemente] / (self.liste_elemente.L_in_mm[index_elemente]) # Elementsteifigkeitsmatrix    
            
            self.liste_elemente.KE[index_elemente] = self.liste_elemente.KE[index_elemente] * temp_E_in_MPa
            
            if (Ausgaben_schreiben == True):
                print('Elementsteifigkeitsmatrix Element '+str(temp_Element_id)+':\n', self.liste_elemente.KE[index_elemente])       
                print('\n')         
                            

datenlogger = class_datenlogger()
bibliothek_materialien = class_bibliothek_materialien()
liste_knoten = class_knoten()
liste_elemente = class_elemente()
class_strucsim_main = class_strucsim_funktionen(bibliothek_materialien, liste_knoten, liste_elemente)